import { getObjectMetadata } from "../common/metadata";
import { GLOBAL_SERVICE_PROVIDER, LOG_CONTEXT } from "../common/constants";
import { ErrorBuilder } from "../system/Errors";
import { AbstractLogger } from "../logging/AbstractLogger";
const isSubClass = (derivedClass, baseClass) => {
    return derivedClass === baseClass || derivedClass.prototype instanceof baseClass;
};
class DefaultServiceInjector {
    injectableRegistry;
    providerRegistry;
    initables = [];
    constructor(injectableRegistry, providerRegistry) {
        this.injectableRegistry = injectableRegistry;
        this.providerRegistry = providerRegistry;
        this.assertNoCircularDependency();
    }
    async initInstances() {
        for (const initable of this.initables) {
            await initable.init();
        }
        getServiceProvider().addInitables(this.initables);
        this.initables = [];
    }
    resolveMultiple(abstractConstructorFn) {
        const providers = this.providerRegistry.filter((reg) => isSubClass(reg.baseClass, abstractConstructorFn));
        const injectables = this.injectableRegistry.filter((reg) => isSubClass(reg.constructorFn, abstractConstructorFn));
        if (!providers.length && !injectables.length) {
            return [];
        }
        //-- prioritize provider
        if (providers.length) {
            return providers.map((providerInfo) => {
                const provider = this.resolve(providerInfo.provider);
                let instance;
                if (providerInfo.mode !== "singleton" || !providerInfo.singletonInstance) {
                    instance = provider.provide();
                    if (providerInfo.mode === "singleton") {
                        providerInfo.singletonInstance = instance;
                    }
                    const metadata = getObjectMetadata(instance.constructor);
                    if (metadata?.initable) {
                        this.initables.push(instance);
                    }
                }
                else {
                    instance = providerInfo.singletonInstance;
                }
                return instance;
            });
        }
        if (injectables.length) {
            return injectables.map((reg) => {
                switch (reg.mode) {
                    case "singleton":
                        if (!reg.singletonInstance) {
                            reg.singletonInstance = this.getNewInstance(reg);
                        }
                        return reg.singletonInstance;
                    case "transient":
                        return this.getNewInstance(reg);
                }
            });
        }
        return [];
    }
    resolveOptional(abstractConstructorFn) {
        const instances = this.resolveMultiple(abstractConstructorFn);
        if (!instances.length) {
            return undefined;
        }
        if (instances.length > 1) {
            throw ErrorBuilder.forMutipleInjectionsFound(abstractConstructorFn, instances.map((instance) => instance.constructor));
        }
        return instances[0];
    }
    resolve(abstractConstructorFn) {
        const instances = this.resolveMultiple(abstractConstructorFn);
        if (!instances.length) {
            throw ErrorBuilder.forInjectionNotFound(abstractConstructorFn);
        }
        if (instances.length > 1) {
            throw ErrorBuilder.forMutipleInjectionsFound(abstractConstructorFn, instances.map((instance) => instance.constructor));
        }
        return instances[0];
    }
    getNewInstance(info) {
        const dependencies = (info.dependencies || []).map((dep) => dep.required ? this.resolve(dep.ctor) : this.resolveOptional(dep.ctor));
        const instance = new info.constructorFn(...dependencies);
        const metadata = getObjectMetadata(info.constructorFn);
        if (metadata?.initable) {
            this.initables.push(instance);
        }
        if (metadata?.logContext) {
            dependencies.forEach((dep) => {
                if (dep instanceof AbstractLogger) {
                    dep[LOG_CONTEXT] = metadata.logContext;
                }
            });
        }
        return instance;
    }
    assertNoCircularDependency() {
        //-- build dag nodes
        let dagNodes = this.injectableRegistry.map((reg) => ({
            constructorFn: reg.constructorFn,
            dependencies: (reg.dependencies || []).slice(),
        }));
        let removeCount = dagNodes.length;
        while (removeCount > 0) {
            //-- remove nodes that has empty dependency until can no more remove
            const tobeRemoved = dagNodes.filter((node) => 
            //-- no dependency
            !node.dependencies.length ||
                //-- or every dependency has no candidate
                node.dependencies.every((dep) => !dagNodes.some((otherNode) => isSubClass(otherNode.constructorFn, dep.ctor))));
            removeCount = tobeRemoved.length;
            //-- remove tobeRemoved nodes from dagNodes
            dagNodes = dagNodes.filter((node) => !tobeRemoved.includes(node));
            //-- remove the tobeRemoved nodes from dependencies list of remaining nodes
            dagNodes.forEach((node) => {
                node.dependencies = node.dependencies.filter((dep) => !tobeRemoved.find((removeNode) => isSubClass(removeNode.constructorFn, dep.ctor)));
            });
        }
        if (dagNodes.length) {
            throw ErrorBuilder.forInjectionCircularDependency(dagNodes.map((dag) => dag.constructorFn));
        }
    }
}
class ServiceProvider {
    injectableRegistry = [];
    providerRegistry = [];
    initables = [];
    //@internal
    addInitables(initables) {
        this.initables = this.initables.concat(initables);
    }
    //@internal
    getInitables() {
        return this.initables;
    }
    getInfo(constructorFn) {
        return this.injectableRegistry.find((reg) => isSubClass(reg.constructorFn, constructorFn));
    }
    register(constructorFn, mode) {
        const index = this.injectableRegistry.findIndex((reg) => reg.constructorFn === constructorFn);
        const reg = {
            constructorFn,
            mode: mode || "singleton",
            dependencies: getObjectMetadata(constructorFn)?.constructorDependencies ||
                (Reflect.getMetadata("design:paramtypes", constructorFn) || []).map((ctor) => ({
                    ctor,
                    required: true,
                })),
        };
        if (index < 0) {
            this.injectableRegistry.push(reg);
        }
        else {
            this.injectableRegistry[index] = { ...this.injectableRegistry[index], ...reg };
        }
        return this;
    }
    addProvider(baseClass, provider) {
        const index = this.providerRegistry.findIndex((reg) => reg.baseClass === baseClass);
        const metadata = getObjectMetadata(provider);
        const reg = { baseClass, provider, mode: metadata?.injectionMode || "singleton" };
        if (index < 0) {
            this.providerRegistry.push(reg);
        }
        else {
            this.providerRegistry[index] = reg;
        }
        return this;
    }
    getInjector() {
        return new DefaultServiceInjector(this.injectableRegistry.slice(), this.providerRegistry.slice());
    }
}
export const getServiceProvider = () => {
    let serviceProvider = globalThis[GLOBAL_SERVICE_PROVIDER];
    if (!serviceProvider) {
        serviceProvider = new ServiceProvider();
        globalThis[GLOBAL_SERVICE_PROVIDER] = serviceProvider;
    }
    return serviceProvider;
};
export const getInstance = async (cls) => {
    const injector = getServiceProvider().getInjector();
    const instance = injector.resolve(cls);
    await injector.initInstances();
    return instance;
};
