import { DataType } from "../common/DataType";
import { initObjectMetadata, initFieldMetadata, getObjectMetadata } from "../common/metadata";
import { Errors } from "../system/Errors";
/**
 * A simple data transfer object.
 */
export const Data = (config) => (constructor) => {
    const metadata = initObjectMetadata(constructor.prototype);
    metadata.relaxed = config?.relaxed;
    metadata.description = config?.description;
    metadata.ignoreCrud = config?.ignoreCrud;
    //-- try create sample model and check for default values
    const sampleModel = new constructor();
    const defaultFields = Object.keys(sampleModel).filter((key) => metadata.fields.find((f) => f.name === key));
    defaultFields.forEach((defaultField) => {
        const field = initFieldMetadata(constructor.prototype, defaultField);
        field.defaultValue = sampleModel[defaultField];
    });
    return constructor;
};
export const Field = (config) => (prototype, propertyKey) => {
    const field = initFieldMetadata(prototype, propertyKey);
    Object.assign(field, config);
    //-- properties interpolation
    if (field.isTimestamp) {
        field.isBigInt = true;
        field.rangeProps = { min: 0 };
    }
    if (field.vectorProps) {
        field.dataType = DataType.ARRAY;
    }
    else if (field.elementDto) {
        field.dataType = DataType.OBJECT;
    }
    //-- properties validation
    if (field.mimeProps && field.dataType !== DataType.STRING) {
        throw Errors.SYSTEM_ERROR(`Cannot use uri on non-text field of ${prototype.constructor.name}:${propertyKey}`);
    }
    if (field.isBigInt && field.dataType !== DataType.NUMBER) {
        throw Errors.SYSTEM_ERROR(`Cannot use big int on non-numeric field of ${prototype.constructor.name}:${propertyKey}`);
    }
    if (field.textLength && (field.textLength <= 0 || !Number.isInteger(field.textLength))) {
        throw Errors.SYSTEM_ERROR(`Text length only accepts integer value, provided: ${field.textLength}`);
    }
};
export const Relation = (model, cascadeType) => {
    const metadata = getObjectMetadata(model);
    if (!metadata) {
        throw Errors.SYSTEM_ERROR(`Cannot find dto metadata: ${model.name}`);
    }
    return { fk: { dtoId: metadata.id, cascade: cascadeType } };
};
export const Enum = (enumClass) => {
    const values = Object.values(enumClass);
    const numberValues = values.filter((v) => typeof v === "number");
    const stringValues = values.filter((v) => typeof v === "string");
    return { enum: numberValues.length ? numberValues : stringValues };
};
export const ElementDto = (dtoClass) => {
    return { elementDto: getObjectMetadata(dtoClass) };
};
const Has = (config) => (prototype, propertyKey) => {
    const dtoMedata = getObjectMetadata(config.model);
    const field = initFieldMetadata(prototype, propertyKey);
    field.vectorProps = config.single ? undefined : { ...field.vectorProps, elementDataType: DataType.OBJECT };
    field.hasMany = {
        relationDto: dtoMedata,
        column: config.column,
        single: config.single,
    };
    //-- init FK for target model
    const targetField = initFieldMetadata(config.model.prototype, config.column);
    targetField.fk = Relation(prototype.constructor, config.cascade || "delete").fk;
};
export const HasOne = (config) => (prototype, propertyKey) => {
    return Has({ ...config, single: true })(prototype, propertyKey);
};
export const HasMany = (config) => (prototype, propertyKey) => {
    return Has({ ...config, single: false })(prototype, propertyKey);
};
