import { DataType } from "../common/DataType";
import { Errors } from "../system/Errors";
export const uniqueReducer = (collection, item) => collection.includes(item) ? collection : collection.concat(item);
export const deepMerge = (target, source, joinArray = false) => {
    if (target === undefined || source === undefined) {
        return;
    }
    if (typeof target !== "object" || typeof source !== "object") {
        return;
    }
    //-- source may be null
    const sourceKeys = Object.keys(source || {});
    for (const key of sourceKeys) {
        const sourceValue = source[key];
        if (sourceValue !== null &&
            typeof sourceValue === "object" &&
            typeof target[key] === "object" &&
            !Array.isArray(sourceValue)) {
            if (!target[key]) {
                target[key] = {};
            }
            deepMerge(target[key], sourceValue);
        }
        else {
            if (Array.isArray(sourceValue) && joinArray) {
                target[key] = [...(target[key] || []), ...sourceValue];
            }
            else if (sourceValue !== undefined) {
                target[key] = sourceValue;
            }
        }
    }
};
export const pickData = (data, fields) => {
    const result = {};
    for (const f of fields) {
        result[f] = data[f] === null ? undefined : data[f];
    }
    return result;
};
export const omitData = (data, fields) => {
    const result = {};
    const allFields = Object.keys(data).filter((f) => !fields.includes(f));
    for (const f of allFields) {
        result[f] = data[f] === null ? undefined : data[f];
    }
    return result;
};
export const diffData = (oldObject, newObject) => {
    if (!oldObject || !newObject) {
        return newObject;
    }
    const keys = Object.keys(newObject).filter((k) => {
        const oldValue = oldObject[k];
        const newValue = newObject[k];
        return !(oldValue === newValue ||
            ((oldValue === null || oldValue === undefined) && newValue === "") ||
            (oldValue === "" && (newValue === null || newValue === undefined)) ||
            (oldValue === undefined && newValue === null) ||
            (oldValue === null && newValue === undefined));
    });
    if (!keys.length) {
        return undefined;
    }
    return keys.reduce((collector, k) => {
        let newValue = newObject[k];
        if (newValue && typeof newValue === "object" && !Array.isArray(newValue)) {
            const oldValue = oldObject[k];
            newValue = diffData(oldValue, newValue);
        }
        return Object.assign(collector, { [k]: newValue });
    }, {});
};
export const leanData = (obj) => {
    return (obj &&
        Object.keys(obj)
            .filter((k) => obj[k] !== undefined)
            .reduce((collector, k) => Object.assign(collector, { [k]: obj[k] }), {}));
};
export const stripData = (value, metadata) => {
    if (metadata.primitiveType) {
        return value;
    }
    const newValue = {};
    for (const field of metadata.fields) {
        let fieldValue = value[field.name];
        if (fieldValue != null && fieldValue != undefined) {
            switch (field.dataType) {
                case DataType.ARRAY:
                    if (field.vectorProps?.elementDataType === DataType.OBJECT && field.elementDto) {
                        //-- check and validate every element in array
                        fieldValue = fieldValue.map((v) => stripData(v, field.elementDto));
                        break;
                    }
                    break;
                case DataType.OBJECT:
                    if (field.elementDto) {
                        fieldValue = stripData(fieldValue, field.elementDto);
                    }
                    break;
            }
        }
        newValue[field.name] = fieldValue;
    }
    return newValue;
};
export const diffNestedArray = (newData, oldData) => {
    const newArray = newData.map((newItem) => {
        if (!newItem.id) {
            //-- new item
            return newItem;
        }
        else {
            //-- diff with original
            const originalItem = oldData.find((p) => p.id === newItem.id);
            const diff = diffData(originalItem, newItem);
            if (!diff) {
                //-- no diff, keep old item
                return { id: newItem.id };
            }
            else {
                //-- update with diff
                return { ...diff, id: newItem.id };
            }
        }
    });
    return newArray;
};
export const validateData = (value, metadata, fillDefault = true) => {
    if (value === undefined || value === null) {
        return value;
    }
    if (metadata.primitiveType) {
        //-- check primitive
        switch (metadata.primitiveType) {
            case DataType.NUMBER:
                if ((!metadata.relaxed && typeof value !== "number") || isNaN(value)) {
                    throw Errors.VALIDATION_ERROR(`Number value is required`);
                }
                return Number(value);
            case DataType.BOOLEAN:
                if (!metadata.relaxed && typeof value !== "boolean") {
                    throw Errors.VALIDATION_ERROR(`Boolean value is required`);
                }
                return value === "true" || Boolean(value);
            case DataType.STRING:
                if (!metadata.relaxed && typeof value !== "string") {
                    throw Errors.VALIDATION_ERROR(`String value is required`);
                }
                return String(value);
        }
    }
    const newValue = {};
    for (const field of metadata.fields) {
        let fieldValue = value[field.name];
        if (fieldValue === undefined && field.defaultValue !== undefined && fillDefault) {
            fieldValue = field.defaultValue;
        }
        if (fieldValue === undefined || (fieldValue === "" && field.dataType === DataType.STRING)) {
            if (field.isRequired && !field.serverValue) {
                throw Errors.VALIDATION_ERROR(`Missing required value of field: ${field.name}`);
            }
        }
        else if (fieldValue === null) {
            if (!field.allowNull) {
                throw Errors.VALIDATION_ERROR(`Null value is not allowed for field: ${field.name}`);
            }
        }
        else {
            switch (field.dataType) {
                case DataType.NUMBER:
                    if ((!metadata.relaxed && typeof fieldValue !== "number") || isNaN(fieldValue)) {
                        throw Errors.VALIDATION_ERROR(`Numeric value is required for field: ${field.name}`);
                    }
                    fieldValue = Number(fieldValue);
                    if (!field.floating && !Number.isInteger(fieldValue)) {
                        throw Errors.VALIDATION_ERROR(`Integer value is required for field: ${field.name}`);
                    }
                    if (field.rangeProps) {
                        if ((field.rangeProps.min !== undefined &&
                            ((field.rangeProps.minExclusive && field.rangeProps.min >= fieldValue) ||
                                (!field.rangeProps.minExclusive && field.rangeProps.min > fieldValue))) ||
                            (field.rangeProps.max !== undefined &&
                                ((field.rangeProps.maxExclusive && field.rangeProps.max <= fieldValue) ||
                                    (!field.rangeProps.maxExclusive && field.rangeProps.max < fieldValue)))) {
                            throw Errors.VALIDATION_ERROR(`Field value of range: ${field.name}`);
                        }
                    }
                    break;
                case DataType.STRING:
                    if (!metadata.relaxed && typeof fieldValue !== "string") {
                        throw Errors.VALIDATION_ERROR(`String value is required for field: ${field.name}`);
                    }
                    fieldValue = String(fieldValue);
                    if (field.regex) {
                        const regex = new RegExp(field.regex);
                        if (!regex.exec(fieldValue)) {
                            throw Errors.VALIDATION_ERROR(`Regex check not pass: ${field.name}`);
                        }
                    }
                    break;
                case DataType.BOOLEAN:
                    if (!metadata.relaxed && typeof fieldValue !== "boolean") {
                        throw Errors.VALIDATION_ERROR(`Boolean value is required for field: ${field.name}`);
                    }
                    fieldValue = fieldValue === "true" || Boolean(fieldValue);
                    break;
                case DataType.ARRAY:
                    if (!metadata.relaxed && !Array.isArray(fieldValue)) {
                        throw Errors.VALIDATION_ERROR(`Array value is required for field: ${field.name}`);
                    }
                    if (field.vectorProps) {
                        if ((field.vectorProps.minLength && fieldValue.length < field.vectorProps.minLength) ||
                            (field.vectorProps.maxLength && fieldValue.length > field.vectorProps.maxLength)) {
                            throw Errors.VALIDATION_ERROR(`Array length not in range for field: ${field.name}`);
                        }
                        if (field.vectorProps.elementDataType) {
                            switch (field.vectorProps.elementDataType) {
                                case DataType.ARRAY:
                                    let index = fieldValue.findIndex((element) => !Array.isArray(element));
                                    if (index >= 0) {
                                        throw Errors.VALIDATION_ERROR(`Element data type mismatched at: ${field.name}[${index}]`);
                                    }
                                    break;
                                case DataType.OBJECT:
                                    //-- check and validate every element in array
                                    fieldValue.forEach((element, index) => {
                                        if (typeof element !== "object") {
                                            throw Errors.VALIDATION_ERROR(`Element data type mismatched at: ${field.name}[${index}]`);
                                        }
                                        if (field.elementDto) {
                                            try {
                                                fieldValue[index] = validateData(element, field.elementDto, fillDefault);
                                            }
                                            catch (err) {
                                                throw Errors.VALIDATION_ERROR(`Validation failed for element at: ${field.name}[${index}], nested exception: [${err.name}] ${err.message}`);
                                            }
                                        }
                                    });
                                    break;
                                default:
                                    if (fieldValue.some((element) => (element === null && !field.vectorProps?.allowNull) ||
                                        (element !== null &&
                                            typeof element !== field.vectorProps?.elementDataType))) {
                                        throw Errors.VALIDATION_ERROR(`Element data type mismatched: ${field.name}`);
                                    }
                                    break;
                            }
                        }
                        if (!field.vectorProps.allowDuplicated &&
                            fieldValue.some((element) => fieldValue.filter((e) => e === element).length > 1)) {
                            throw Errors.VALIDATION_ERROR(`Duplicated values found in field: ${field.name}`);
                        }
                        if (field.vectorProps.superSet &&
                            fieldValue.some((element) => !field.vectorProps?.superSet?.includes(element))) {
                            throw Errors.VALIDATION_ERROR(`Found not allowed value in field: ${field.name}`);
                        }
                    }
                    break;
                case DataType.OBJECT:
                    if (field.elementDto) {
                        //-- recursively call validateDTO to validate inner field
                        try {
                            fieldValue = validateData(fieldValue, field.elementDto, fillDefault);
                        }
                        catch (err) {
                            throw Errors.VALIDATION_ERROR(`Error validating: "${field.name}", nested exception: [${err.name} ${err.message}]`);
                        }
                    }
                    else if (field.mimeProps) {
                        //-- check mime
                        const file = fieldValue;
                        if (file && field.mimeProps.maxFileKB && file.size > field.mimeProps.maxFileKB * 1024) {
                            throw Errors.VALIDATION_ERROR(`Max file size exceeded: ${(file.size / 1024).toFixed()} KB > ${field.mimeProps.maxFileKB} KB`);
                        }
                    }
                    break;
            }
            //-- other check
            if (field.enum && !field.enum.includes(fieldValue)) {
                throw Errors.VALIDATION_ERROR(`Value not belongs to enum in field: ${field.name}`);
            }
        }
        if (fieldValue !== undefined) {
            newValue[field.name] = fieldValue;
        }
    }
    //-- if strict then extra key provided in value is not allowed
    const newValueKeys = Object.keys(newValue);
    const extraKeys = Object.keys(value).filter((key) => !newValueKeys.includes(key) && !metadata.fields.some((f) => f.name === key));
    if (extraKeys.length) {
        if (!metadata.relaxed) {
            throw Errors.VALIDATION_ERROR(`Keys not allowed: ${JSON.stringify(extraKeys)}`);
        }
        else {
            //-- copy extra key to new value
            extraKeys.forEach((key) => {
                newValue[key] = value[key];
            });
        }
    }
    return newValue;
};
