import { initFieldMetadata, initObjectMetadata, getObjectMetadata, getTableName } from "../common/metadata";
import { getGlobalStore } from "../system/system";
import { Data } from "../dto/decorator";
import { Errors } from "../system/Errors";
const registerModel = (model) => {
    const store = getGlobalStore();
    if (!store.models) {
        store.models = [];
    }
    const models = store.models;
    const modelMetadata = getObjectMetadata(model);
    const duplicatedModelIndex = models.findIndex((m) => {
        const metadata = getObjectMetadata(m);
        return metadata?.id === modelMetadata?.id;
    });
    if (duplicatedModelIndex >= 0) {
        models[duplicatedModelIndex] = model;
    }
    else {
        //-- check if this model has @HasMany on it, if so find the appropriate position
        const hasManyFields = modelMetadata?.fields.filter((f) => f.hasMany) || [];
        if (!hasManyFields.length) {
            //-- no has many fields, just add to array because it complies with DAG order
            models.push(model);
        }
        else {
            //-- this model contains other models as @HasMany / @HasOne, find the position of the first child model and put it right before
            const firstChildPosition = models.findIndex((m) => {
                const childMeta = getObjectMetadata(m);
                return hasManyFields.find((f) => f.hasMany?.relationDto.id === childMeta?.id);
            });
            if (firstChildPosition >= 0) {
                //-- insert to array at firstChildPosition
                models.splice(firstChildPosition, 0, model);
            }
            else {
                //-- just insert it normally
                models.push(model);
            }
        }
    }
};
export const Model = (config) => (constructor) => {
    const meta = initObjectMetadata(constructor.prototype);
    meta.id = getTableName(config?.override || constructor);
    //-- register model here
    registerModel(constructor);
    return Data({ description: config?.description, ignoreCrud: config?.ignoreCrud })(constructor);
};
export const Index = (fields, config) => (constructor) => {
    const meta = initObjectMetadata(constructor.prototype);
    if (!meta.indexes) {
        meta.indexes = [];
    }
    meta.indexes.push({
        columns: fields,
        sparse: config?.sparse,
        unique: config?.unique,
    });
};
export const Column = (props) => (prototype, propertyKey) => {
    const field = initFieldMetadata(prototype, propertyKey);
    Object.assign(field, props);
    if (field.pk) {
        field.isRequired = true;
        if (field.pk.isAutoIncrement) {
            field.serverValue = true;
        }
    }
    if (field.creationTimestamp || field.modificationTimestamp) {
        field.isTimestamp = true;
        field.serverValue = true;
        field.isRequired = true;
    }
    if (field.isTimestamp) {
        field.isBigInt = true;
    }
    if (field.isSymbol && !field.textLength) {
        throw Errors.SYSTEM_ERROR(`Symbol is specified without fixed text length ${prototype.constructor.name}:${propertyKey}`);
    }
    //-- properties validation
    if (field.pk && field.fk) {
        throw Errors.SYSTEM_ERROR(`Field both declared as primary key and foreign key ${prototype.constructor.name}:${propertyKey}`);
    }
};
export const FK = (config) => {
    //-- get primary key of foreign model
    let props = {};
    let relationDto;
    if (config.model) {
        relationDto = getObjectMetadata(config.model);
        const primaryKey = relationDto.fields.find((field) => field.pk);
        if (primaryKey) {
            props.isBigInt = primaryKey.isBigInt;
            props.textLength = primaryKey.textLength;
        }
    }
    props.fk = {
        dtoId: relationDto?.id,
        cascade: config.cascade,
    };
    return props;
};
